/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, version 1.0 RC 1        *
*                (c) 2006-2011 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                              SOFA :: Framework                              *
*                                                                             *
* Authors: The SOFA Team (see Authors.txt)                                    *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/
#ifndef SOFA_DEFAULTTYPE_DATATYPEINFO_H
#define SOFA_DEFAULTTYPE_DATATYPEINFO_H

#include <vector>
#include <sofa/helper/fixed_array.h>
//#include <sofa/helper/vector.h>
#include <sofa/helper/set.h>
#include <sstream>

namespace sofa
{

namespace helper
{
template <class T, class MemoryManager >
class vector;
}

namespace defaulttype
{

/** Type traits class for objects stored in Data

    DataTypeInfo is part of the introspection/reflection capabilities of the
    Sofa scene graph API; it is used to manipulate Data values generically
    in template code, and work with different types of containers, integers,
    scalars (float, double...), strings, etc

    For example, it can be used to work with arrays without having to handle all
    the possible array classes used in Sofa: fixed or dynamic size, CPU or GPU,
    etc.
*/
template<class TDataType>
struct DataTypeInfo
{
    typedef TDataType DataType;
    typedef DataType BaseType; ///< class of type contained in DataType
    typedef DataType ValueType; ///< type of the final atomic values
    typedef DataTypeInfo<BaseType> BaseTypeInfo;
    typedef DataTypeInfo<ValueType> ValueTypeInfo;

    enum { ValidInfo       = 0 }; ///< 1 if this type has valid infos
    enum { FixedSize       = 0 }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 0 }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 0 }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 0 }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = 0 }; ///< 1 if this type uses integer values
    enum { Scalar          = 0 }; ///< 1 if this type uses scalar values
    enum { Text            = 0 }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 0 }; ///< 1 if this type uses copy-on-write

    enum { Size = 1 }; ///< largest known fixed size for this type, as returned by size()
    static size_t size() { return 1; }

    static size_t size(const DataType& /*data*/) { return 1; }

    template <typename T>
    static void getValue(const DataType& /*data*/, size_t /*index*/, T& /*value*/)
    {
    }

    static void setSize(DataType& /*data*/, size_t /*size*/) {  }

    template<typename T>
    static void setValue(DataType& /*data*/, size_t /*index*/, const T& /*value*/)
    {
    }

    static void getValueString(const DataType& /*data*/, size_t /*index*/, std::string& /*value*/)
    {
    }

    static void setValueString(DataType& /*data*/, size_t /*index*/, const std::string& /*value*/)
    {
    }

    static const char* name() { return "unknown"; }

};

/// Type name template: default to using DataTypeInfo::name(), but can be overriden for types with shorter typedefs
template<class TDataType>
struct DataTypeName : public DataTypeInfo<TDataType>
{
};


/** Abstract type traits class for objects stored in Data

    AbstractTypeInfo is part of the introspection/reflection capabilities of the
    Sofa scene graph API; it is used to manipulate Data values generically when
    their type is unknown, and to get information about their type (Is it a
    primitive type ? A string ? How much memory should be allocated to copy it?)

    Use BaseData::getValueTypeInfo() to get a pointer to an AbtractTypeInfo.

    This class is primarily used to copy information accross BaseData, for
    example when there exists a link between two instances of BaseData.
    E.g. this mecanism allows you to copy the content of a Data<vector<int>>
    into a Data<vector<double>>, because there is an acceptable conversion
    between integer and double, and because both Data use a resizable container.
*/
class AbstractTypeInfo
{
public:
    virtual const AbstractTypeInfo* BaseType() const = 0;
    virtual const AbstractTypeInfo* ValueType() const = 0;

    virtual std::string name() const = 0;

    virtual bool ValidInfo() const = 0;
    virtual bool FixedSize() const = 0;
    virtual bool ZeroConstructor() const = 0;
    virtual bool SimpleCopy() const = 0;
    virtual bool SimpleLayout() const = 0;
    virtual bool Integer() const = 0;
    virtual bool Scalar() const = 0;
    virtual bool Text() const = 0;
    virtual bool CopyOnWrite() const = 0;

    virtual size_t size() const = 0;
    virtual size_t size(const void* data) const = 0;
    virtual void setSize(void* data, size_t size) const = 0;

    virtual long long   getIntegerValue(const void* data, size_t index) const = 0;
    virtual double      getScalarValue (const void* data, size_t index) const = 0;
    virtual std::string getTextValue   (const void* data, size_t index) const = 0;

    virtual void setIntegerValue(void* data, size_t index, long long value) const = 0;
    virtual void setScalarValue (void* data, size_t index, double value) const = 0;
    virtual void setTextValue(void* data, size_t index, const std::string& value) const = 0;

protected: // only derived types can instantiate this class
    AbstractTypeInfo() {}
    virtual ~AbstractTypeInfo() {}

private: // copy constructor or operator forbidden
    AbstractTypeInfo(const AbstractTypeInfo&) {}
    void operator=(const AbstractTypeInfo&) {}
};

/// Abstract type traits class
template<class TDataType>
class VirtualTypeInfo : public AbstractTypeInfo
{
public:
    typedef TDataType DataType;
    typedef DataTypeInfo<DataType> Info;

    static VirtualTypeInfo* get() { static VirtualTypeInfo<DataType> t; return &t; }

    virtual const AbstractTypeInfo* BaseType() const  { return VirtualTypeInfo<typename Info::BaseType>::get(); }
    virtual const AbstractTypeInfo* ValueType() const { return VirtualTypeInfo<typename Info::ValueType>::get(); }

    virtual std::string name() const { return DataTypeName<DataType>::name(); }

    virtual bool ValidInfo() const       { return Info::ValidInfo; }
    virtual bool FixedSize() const       { return Info::FixedSize; }
    virtual bool ZeroConstructor() const { return Info::ZeroConstructor; }
    virtual bool SimpleCopy() const      { return Info::SimpleCopy; }
    virtual bool SimpleLayout() const    { return Info::SimpleLayout; }
    virtual bool Integer() const         { return Info::Integer; }
    virtual bool Scalar() const          { return Info::Scalar; }
    virtual bool Text() const            { return Info::Text; }
    virtual bool CopyOnWrite() const     { return Info::CopyOnWrite; }

    virtual size_t size() const
    {
        return Info::size();
    }
    virtual size_t size(const void* data) const
    {
        return Info::size(*(const DataType*)data);
    }
    virtual void setSize(void* data, size_t size) const
    {
        Info::setSize(*(DataType*)data, size);
    }

    virtual long long getIntegerValue(const void* data, size_t index) const
    {
        long long v = 0;
        Info::getValue(*(const DataType*)data, index, v);
        return v;
    }

    virtual double    getScalarValue (const void* data, size_t index) const
    {
        double v = 0;
        Info::getValue(*(const DataType*)data, index, v);
        return v;
    }

    virtual std::string getTextValue   (const void* data, size_t index) const
    {
        std::string v;
        Info::getValueString(*(const DataType*)data, index, v);
        return v;
    }

    virtual void setIntegerValue(void* data, size_t index, long long value) const
    {
        Info::setValue(*(DataType*)data, index, value);
    }

    virtual void setScalarValue (void* data, size_t index, double value) const
    {
        Info::setValue(*(DataType*)data, index, value);
    }

    virtual void setTextValue(void* data, size_t index, const std::string& value) const
    {
        Info::setValueString(*(DataType*)data, index, value);
    }

protected: // only derived types can instantiate this class
    VirtualTypeInfo() {}
};

template<class TDataType>
struct IntegerTypeInfo
{
    typedef TDataType DataType;
    typedef DataType BaseType;
    typedef DataType ValueType;
    typedef long long ConvType; ///< preferred type for conversions (i.e. long long for integers, double for scalars)
    typedef IntegerTypeInfo<DataType> BaseTypeInfo;
    typedef IntegerTypeInfo<DataType> ValueTypeInfo;

    enum { ValidInfo       = 1 }; ///< 1 if this type has valid infos
    enum { FixedSize       = 1 }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 1 }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 1 }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 1 }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = 1 }; ///< 1 if this type uses integer values
    enum { Scalar          = 0 }; ///< 1 if this type uses scalar values
    enum { Text            = 0 }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 0 }; ///< 1 if this type uses copy-on-write

    enum { Size = 1 }; ///< largest known fixed size for this type, as returned by size()
    static size_t size() { return 1; }

    static size_t size(const DataType& /*data*/) { return 1; }

    static void setSize(DataType& /*data*/, size_t /*size*/) {  }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (index != 0) return;
        value = static_cast<T>(data);
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (index != 0) return;
        data = static_cast<DataType>(value);
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (index != 0) return;
        std::ostringstream o; o << data; value = o.str();
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (index != 0) return;
        std::istringstream i(value); i >> data;
    }
};

struct BoolTypeInfo
{
    typedef bool DataType;
    typedef DataType BaseType;
    typedef DataType ValueType;
    typedef long long ConvType; ///< preferred type for conversions (i.e. long long for integers, double for scalars)
    typedef IntegerTypeInfo<DataType> BaseTypeInfo;
    typedef IntegerTypeInfo<DataType> ValueTypeInfo;

    enum { ValidInfo       = 1 }; ///< 1 if this type has valid infos
    enum { FixedSize       = 1 }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 1 }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 1 }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 1 }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = 1 }; ///< 1 if this type uses integer values
    enum { Scalar          = 0 }; ///< 1 if this type uses scalar values
    enum { Text            = 0 }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 0 }; ///< 1 if this type uses copy-on-write

    enum { Size = 1 }; ///< largest known fixed size for this type, as returned by size()
    static size_t size() { return 1; }

    static size_t size(const DataType& /*data*/) { return 1; }

    static void setSize(DataType& /*data*/, size_t /*size*/) {  }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (index != 0) return;
        value = static_cast<T>(data);
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (index != 0) return;
        data = (value != 0);
    }

    template<typename T>
    static void setValue(std::vector<DataType>::reference data, size_t index, const T& v )
    {
        if (index != 0) return;
        data = (v != 0);
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (index != 0) return;
        std::ostringstream o; o << data; value = o.str();
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (index != 0) return;
        std::istringstream i(value); i >> data;
    }

    static void setValueString(std::vector<DataType>::reference data, size_t index, const std::string& value )
    {
        if (index != 0) return;
        bool b = data;
        std::istringstream i(value); i >> b;
        data = b;
    }
};

template<class TDataType>
struct ScalarTypeInfo
{
    typedef TDataType DataType;
    typedef DataType BaseType;
    typedef DataType ValueType;
    typedef long long ConvType; ///< preferred type for conversions (i.e. long long for integers, double for scalars)
    typedef ScalarTypeInfo<TDataType> BaseTypeInfo;
    typedef ScalarTypeInfo<TDataType> ValueTypeInfo;

    enum { ValidInfo       = 1 }; ///< 1 if this type has valid infos
    enum { FixedSize       = 1 }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 1 }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 1 }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 1 }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = 0 }; ///< 1 if this type uses integer values
    enum { Scalar          = 1 }; ///< 1 if this type uses scalar values
    enum { Text            = 0 }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 0 }; ///< 1 if this type uses copy-on-write

    enum { Size = 1 }; ///< largest known fixed size for this type, as returned by size()
    static size_t size() { return 1; }

    static size_t size(const DataType& /*data*/) { return 1; }

    static void setSize(DataType& /*data*/, size_t /*size*/) {  }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (index != 0) return;
        value = static_cast<T>(data);
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (index != 0) return;
        data = static_cast<DataType>(value);
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (index != 0) return;
        std::ostringstream o; o << data; value = o.str();
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (index != 0) return;
        std::istringstream i(value); i >> data;
    }
};

template<class TDataType>
struct TextTypeInfo
{
    typedef TDataType DataType;
    typedef DataType BaseType;
    typedef DataType ValueType;
    typedef long long ConvType; ///< preferred type for conversions (i.e. long long for integers, double for scalars)
    typedef ScalarTypeInfo<TDataType> BaseTypeInfo;
    typedef ScalarTypeInfo<TDataType> ValueTypeInfo;

    enum { ValidInfo       = 1 }; ///< 1 if this type has valid infos
    enum { FixedSize       = 0 }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 0 }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 0 }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 0 }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = 0 }; ///< 1 if this type uses integer values
    enum { Scalar          = 0 }; ///< 1 if this type uses scalar values
    enum { Text            = 1 }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 1 }; ///< 1 if this type uses copy-on-write

    enum { Size = 1 }; ///< largest known fixed size for this type, as returned by size()
    static size_t size() { return 1; }

    static size_t size(const DataType& /*data*/) { return 1; }

    static void setSize(DataType& /*data*/, size_t /*size*/) {  }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (index != 0) return;
        std::istringstream i(data); i >> value;
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (index != 0) return;
        std::ostringstream o; o << value; data = o.str();
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (index != 0) return;
        value = data;
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (index != 0) return;
        data = value;
    }
};

template<class TDataType, int static_size = TDataType::static_size>
struct FixedArrayTypeInfo
{
    typedef TDataType DataType;
    typedef typename DataType::size_type size_type;
    typedef typename DataType::value_type BaseType;
    typedef DataTypeInfo<BaseType> BaseTypeInfo;
    typedef typename BaseTypeInfo::ValueType ValueType;
    typedef DataTypeInfo<ValueType> ValueTypeInfo;

    enum { ValidInfo       = BaseTypeInfo::ValidInfo       }; ///< 1 if this type has valid infos
    enum { FixedSize       = BaseTypeInfo::FixedSize       }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = BaseTypeInfo::ZeroConstructor }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = BaseTypeInfo::SimpleCopy      }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = BaseTypeInfo::SimpleLayout    }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = BaseTypeInfo::Integer         }; ///< 1 if this type uses integer values
    enum { Scalar          = BaseTypeInfo::Scalar          }; ///< 1 if this type uses scalar values
    enum { Text            = BaseTypeInfo::Text            }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 1                             }; ///< 1 if this type uses copy-on-write

    enum { Size = static_size * BaseTypeInfo::Size }; ///< largest known fixed size for this type, as returned by size()
    static size_t size()
    {
        return DataType::size() * BaseTypeInfo::size();
    }

    static size_t size(const DataType& data)
    {
        if (FixedSize)
            return size();
        else
        {
            size_t s = 0;
            for (size_t i=0; i<DataType::size(); ++i)
                s+= BaseTypeInfo::size(data[(size_type)i]);
            return s;
        }
    }

    static void setSize(DataType& data, size_t size)
    {
        if (!FixedSize)
        {
            size /= DataType::size();
            for (size_t i=0; i<DataType::size(); ++i)
                BaseTypeInfo::setSize(data[(size_type)i], size);
        }
    }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::getValue(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::getValue(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<DataType::size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::getValue(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::setValue(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::setValue(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<DataType::size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::setValue(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::getValueString(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::getValueString(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<DataType::size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::getValueString(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::setValueString(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::setValueString(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<DataType::size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::setValueString(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }
};

template<class TDataType>
struct VectorTypeInfo
{
    typedef TDataType DataType;
    typedef typename DataType::size_type size_type;
    typedef typename DataType::value_type BaseType;
    typedef DataTypeInfo<BaseType> BaseTypeInfo;
    typedef typename BaseTypeInfo::ValueType ValueType;
    typedef DataTypeInfo<ValueType> ValueTypeInfo;

    enum { ValidInfo       = BaseTypeInfo::ValidInfo       }; ///< 1 if this type has valid infos
    enum { FixedSize       = 0                             }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 0                             }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 0                             }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = BaseTypeInfo::SimpleLayout    }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = BaseTypeInfo::Integer         }; ///< 1 if this type uses integer values
    enum { Scalar          = BaseTypeInfo::Scalar          }; ///< 1 if this type uses scalar values
    enum { Text            = BaseTypeInfo::Text            }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 1                             }; ///< 1 if this type uses copy-on-write

    enum { Size = BaseTypeInfo::Size }; ///< largest known fixed size for this type, as returned by size()
    static size_t size()
    {
        return BaseTypeInfo::size();
    }

    static size_t size(const DataType& data)
    {
        if (BaseTypeInfo::FixedSize)
            return data.size()*BaseTypeInfo::size();
        else
        {
            size_t n = data.size();
            size_t s = 0;
            for (size_t i=0; i<n; ++i)
                s+= BaseTypeInfo::size(data[(size_type)i]);
            return s;
        }
    }

    static void setSize(DataType& data, size_t size)
    {
        if (BaseTypeInfo::FixedSize)
            data.resize(size/BaseTypeInfo::size());
    }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::getValue(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::getValue(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<data.size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::getValue(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    template<typename T>
    static void setValue(DataType &data, size_t index, const T& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::setValue(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::setValue(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<data.size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::setValue(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::getValueString(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::getValueString(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<data.size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::getValueString(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    static void setValueString(DataType &data, size_t index, const std::string& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseTypeInfo::setValueString(data[(size_type)index], 0, value);
        }
        else if (BaseTypeInfo::FixedSize)
        {
            BaseTypeInfo::setValueString(data[(size_type)(index/BaseTypeInfo::size())], index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (size_t i=0; i<data.size(); ++i)
            {
                size_t n = BaseTypeInfo::size(data[(size_type)i]);
                if (index < s+n)
                {
                    BaseTypeInfo::setValueString(data[(size_type)i], index-s, value);
                    break;
                }
                s += n;
            }
        }
    }
};


template<class TDataType>
struct SetTypeInfo
{
    typedef TDataType DataType;
    typedef typename DataType::value_type BaseType;
    typedef DataTypeInfo<BaseType> BaseTypeInfo;
    typedef typename BaseTypeInfo::ValueType ValueType;
    typedef DataTypeInfo<ValueType> ValueTypeInfo;

    enum { ValidInfo       = BaseTypeInfo::ValidInfo       }; ///< 1 if this type has valid infos
    enum { FixedSize       = 0                             }; ///< 1 if this type has a fixed size
    enum { ZeroConstructor = 0                             }; ///< 1 if the constructor is equivalent to setting memory to 0
    enum { SimpleCopy      = 0                             }; ///< 1 if copying the data can be done with a memcpy
    enum { SimpleLayout    = 0                             }; ///< 1 if the layout in memory is simply N values of the same base type
    enum { Integer         = BaseTypeInfo::Integer         }; ///< 1 if this type uses integer values
    enum { Scalar          = BaseTypeInfo::Scalar          }; ///< 1 if this type uses scalar values
    enum { Text            = BaseTypeInfo::Text            }; ///< 1 if this type uses text values
    enum { CopyOnWrite     = 1                             }; ///< 1 if this type uses copy-on-write

    enum { Size = BaseTypeInfo::Size }; ///< largest known fixed size for this type, as returned by size()
    static size_t size()
    {
        return BaseTypeInfo::size();
    }

    static size_t size(const DataType& data)
    {
        if (BaseTypeInfo::FixedSize)
            return data.size()*BaseTypeInfo::size();
        else
        {
            size_t s = 0;
            for (typename DataType::const_iterator it = data.begin(), end=data.end(); it!=end; ++it)
                s+= BaseTypeInfo::size(*it);
            return s;
        }
    }

    static void setSize(DataType& data, size_t /*size*/)
    {
        data.clear(); // we can't "resize" a set, so the only meaningfull operation is to clear it, as values will be added dynamically in setValue
    }

    template <typename T>
    static void getValue(const DataType &data, size_t index, T& value)
    {
        if (BaseTypeInfo::FixedSize)
        {
            typename DataType::const_iterator it = data.begin();
            for (size_t i=0; i<index/BaseTypeInfo::size(); ++i) ++it;
            BaseTypeInfo::getValue(*it, index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (typename DataType::const_iterator it = data.begin(), end=data.end(); it!=end; ++it)
            {
                size_t n = BaseTypeInfo::size(*it);
                if (index < s+n)
                {
                    BaseTypeInfo::getValue(*it, index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    template<typename T>
    static void setValue(DataType &data, size_t /*index*/, const T& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseType t;
            BaseTypeInfo::setValue(t, 0, value);
            data.insert(t);
        }
        else
        {
            std::cerr << "ERROR: SetTypeInfo::setValue not implemented for set with composite values." << std::endl;
        }
    }

    static void getValueString(const DataType &data, size_t index, std::string& value)
    {
        if (BaseTypeInfo::FixedSize)
        {
            typename DataType::const_iterator it = data.begin();
            for (size_t i=0; i<index/BaseTypeInfo::size(); ++i) ++it;
            BaseTypeInfo::getValueString(*it, index%BaseTypeInfo::size(), value);
        }
        else
        {
            size_t s = 0;
            for (typename DataType::const_iterator it = data.begin(), end=data.end(); it!=end; ++it)
            {
                size_t n = BaseTypeInfo::size(*it);
                if (index < s+n)
                {
                    BaseTypeInfo::getValueString(*it, index-s, value);
                    break;
                }
                s += n;
            }
        }
    }

    static void setValueString(DataType &data, size_t /*index*/, const std::string& value )
    {
        if (BaseTypeInfo::FixedSize && BaseTypeInfo::size() == 1)
        {
            BaseType t;
            BaseTypeInfo::setValueString(t, 0, value);
            data.insert(t);
        }
        else
        {
            std::cerr << "ERROR: SetTypeInfo::setValueString not implemented for set with composite values." << std::endl;
        }
    }
};

template<>
struct DataTypeInfo<bool> : public BoolTypeInfo
{
    static const char* name() { return "bool"; }
};

template<>
struct DataTypeInfo<char> : public IntegerTypeInfo<char>
{
    static const char* name() { return "char"; }
};

template<>
struct DataTypeInfo<unsigned char> : public IntegerTypeInfo<unsigned char>
{
    static const char* name() { return "unsigned char"; }
};

template<>
struct DataTypeInfo<short> : public IntegerTypeInfo<short>
{
    static const char* name() { return "short"; }
};

template<>
struct DataTypeInfo<unsigned short> : public IntegerTypeInfo<unsigned short>
{
    static const char* name() { return "unsigned short"; }
};

template<>
struct DataTypeInfo<int> : public IntegerTypeInfo<int>
{
    static const char* name() { return "int"; }
};

template<>
struct DataTypeInfo<unsigned int> : public IntegerTypeInfo<unsigned int>
{
    static const char* name() { return "unsigned int"; }
};

template<>
struct DataTypeInfo<long> : public IntegerTypeInfo<long>
{
    static const char* name() { return "long"; }
};

template<>
struct DataTypeInfo<unsigned long> : public IntegerTypeInfo<unsigned long>
{
    static const char* name() { return "unsigned long"; }
};

template<>
struct DataTypeInfo<long long> : public IntegerTypeInfo<long long>
{
    static const char* name() { return "long long"; }
};

template<>
struct DataTypeInfo<unsigned long long> : public IntegerTypeInfo<unsigned long long>
{
    static const char* name() { return "unsigned long long"; }
};

template<>
struct DataTypeInfo<float> : public ScalarTypeInfo<float>
{
    static const char* name() { return "float"; }
};

template<>
struct DataTypeInfo<double> : public ScalarTypeInfo<double>
{
    static const char* name() { return "double"; }
};

template<>
struct DataTypeInfo<std::string> : public TextTypeInfo<std::string>
{
    static const char* name() { return "string"; }
};

template<class T, std::size_t N>
struct DataTypeInfo< sofa::helper::fixed_array<T,N> > : public FixedArrayTypeInfo<sofa::helper::fixed_array<T,N> >
{
    static std::string name() { std::ostringstream o; o << "fixed_array<" << DataTypeName<T>::name() << "," << N << ">"; return o.str(); }
};

template<class T, class Alloc>
struct DataTypeInfo< std::vector<T,Alloc> > : public VectorTypeInfo<std::vector<T,Alloc> >
{
    static std::string name() { std::ostringstream o; o << "std::vector<" << DataTypeName<T>::name() << ">"; return o.str(); }
};

template<class T, class Alloc>
struct DataTypeInfo< sofa::helper::vector<T,Alloc> > : public VectorTypeInfo<sofa::helper::vector<T,Alloc> >
{
    static std::string name() { std::ostringstream o; o << "vector<" << DataTypeName<T>::name() << ">"; return o.str(); }
};

template<class T, class Compare, class Alloc>
struct DataTypeInfo< std::set<T,Compare,Alloc> > : public SetTypeInfo<std::set<T,Compare,Alloc> >
{
    static std::string name() { std::ostringstream o; o << "std::set<" << DataTypeName<T>::name() << ">"; return o.str(); }
};

template<class T, class Compare, class Alloc>
struct DataTypeInfo< helper::set<T,Compare,Alloc> > : public SetTypeInfo<helper::set<T,Compare,Alloc> >
{
    static std::string name() { std::ostringstream o; o << "set<" << DataTypeName<T>::name() << ">"; return o.str(); }
};

} // namespace defaulttype

} // namespace sofa

#endif  // SOFA_DEFAULTTYPE_DATATYPEINFO_H
